package com.tzld.piaoquan.abtest.service;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.base.Strings;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.abtest.client.model.GetGroupRequest;
import com.tzld.piaoquan.abtest.client.model.GetGroupResponse;
import com.tzld.piaoquan.abtest.client.model.Result;
import com.tzld.piaoquan.abtest.model.ExpConfig;
import com.tzld.piaoquan.abtest.util.CommonCollectionUtils;
import com.tzld.piaoquan.abtest.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.math.NumberUtils;
import org.apache.commons.lang3.RandomUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.net.InetAddress;
import java.util.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class GroupService {

    @Autowired
    @Qualifier("redisTemplate")
    private RedisTemplate<String, String> redisTemplate;

    @ApolloJsonValue("${new.exp,config.v2:[]}")
    private List<ExpConfig> expConfigs;

    private final String exp_param_date = "date";
    private final String exp_param_group = "group";
    private final String exp_param_isFromCard = "isFromCard";
    private final String exp_param_chain = "chain";

    private final String scene_open = "open";
    private final String scene_daily_share_return = "daily_share_return";
    private final String scene_not_daily_share_return = "not_daily_share_return";


    public GetGroupResponse getGroup(GetGroupRequest request) {
        try {


            Map<String, List<ExpConfig.Layer>> layerMap = CommonCollectionUtils.toMap(expConfigs, c -> c.getAppType(),
                    c -> c.getLayers());
            List<ExpConfig.Layer> layers = layerMap.get(String.valueOf(request.getAppType()));

            if (CollectionUtils.isEmpty(layers)) {
                return GetGroupResponse.newBuilder()
                        .setResult(Result.newBuilder().setCode(1))
                        .setGroup(Strings.nullToEmpty(request.getGroup()))
                        .build();
            }

            //TODO 1. 确定场景 主动打开、当日分享、历史分享
            String scene = "";
            Map<String, String> expMap = JSONUtils.fromJson(request.getGroup(), new TypeToken<Map<String, String>>() {
            }, Collections.emptyMap());
            long time = NumberUtils.toLong(expMap.get(exp_param_date), 0);
            String isFromCard = expMap.getOrDefault(exp_param_isFromCard, "0");
            if (org.apache.commons.lang.StringUtils.equals(isFromCard, "1")) {
                if (isToday(time)) {
                    // 当日分享
                    scene = scene_daily_share_return;
                } else {
                    // 历史分享
                    scene = scene_not_daily_share_return;
                }
            } else {
                // 主动打开
                scene = scene_open;
            }

            log.info("ExperimentInterceptor scene {}", scene);

            //TODO 2. 根据场景 确定策略并分组
            Map<String, Object> groupMap = new HashMap<>();
            Map<String, String> oldGroupMap = org.apache.commons.lang.StringUtils.equals(expMap.get(exp_param_group), "ab100")
                    ? Collections.emptyMap()
                    : JSONUtils.fromJson(expMap.get(exp_param_group), new TypeToken<Map<String, String>>() {
            }, Collections.emptyMap());
            for (ExpConfig.Layer layer : layers) {
                int bucketNum = layer.getBucketNum() <= 0 ? 100 : layer.getBucketNum();
                Map<String, String> groupRule = layer.getGroupRule();
                if (MapUtils.isEmpty(groupRule)) {
                    // random
                    groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                } else {
                    String policy = groupRule.get(scene);
                    if (org.apache.commons.lang.StringUtils.isBlank(policy)) {
                        // random
                        groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                    }
                    switch (policy) {

                        case "inherit":
                            groupMap.put(layer.getLayerId(), oldGroupMap.getOrDefault(layer.getLayerId(),
                                    RandomUtils.nextInt(0, bucketNum) + ""));
                            break;
                        case "random_once":
                            if (!isToday(time)) {
                                groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                            } else {
                                groupMap.put(layer.getLayerId(), oldGroupMap.getOrDefault(layer.getLayerId(),
                                        RandomUtils.nextInt(0, bucketNum) + ""));
                            }
                            break;
                        case "random_every":
                        default:
                            groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum) + "");
                            break;
                    }
                }
            }

            // 3. 写入header
            Map<String, Object> map = new HashMap<>();
            map.put(exp_param_date, System.currentTimeMillis());
            map.put(exp_param_group, JSONUtils.toJson(groupMap));
            map.put(exp_param_isFromCard, isFromCard);
            map.put(exp_param_chain, org.apache.commons.lang.StringUtils.equals(isFromCard, "1")
                    ? expMap.get(exp_param_chain)
                    : genChainId());

            return GetGroupResponse.newBuilder()
                    .setResult(Result.newBuilder().setCode(1))
                    .setGroup(JSONUtils.toJson(map))
                    .build();

        } catch (Exception e) {
            log.error("get group error", e);
            return GetGroupResponse.newBuilder()
                    .setResult(Result.newBuilder().setCode(0).setMessage("get group error"))
                    .build();
        }
    }


    public static String genChainId() {
        StringBuilder traceIdBuilder = new StringBuilder();
        InetAddress address = InetAddress.getLoopbackAddress();
        if (address == null) {
            return "default";
        }
        String addressStr = address.getHostAddress();
        if (org.apache.commons.lang3.StringUtils.isBlank(addressStr)) {
            return "default";
        }
        String[] addressArr = addressStr.split("\\.");
        if (addressArr.length != 4) {
            return "default";
        }

        traceIdBuilder.append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[0])), 2, '0'))
                .append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[1])), 2, '0'))
                .append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[2])), 2, '0'))
                .append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[3])), 2, '0'))
                .append(System.currentTimeMillis())
                .append(RandomStringUtils.randomAlphabetic(4))
                .append(Thread.currentThread().getId());
        return traceIdBuilder.toString();
    }

    private boolean isToday(long time) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(time);
        Calendar current = Calendar.getInstance();
        current.setTime(new Date());
        return calendar.get(Calendar.YEAR) == current.get(Calendar.YEAR)
                && calendar.get(Calendar.MONTH) == current.get(Calendar.MONTH)
                && calendar.get(Calendar.DAY_OF_MONTH) == current.get(Calendar.DAY_OF_MONTH);
    }
}
